# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions to set baseline PHPP windows."""

from PHX.PHPP import phpp_app

from ph_baseliner.codes.model import BaselineCode
from ph_baseliner.codes.options import ClimateZones
from ph_baseliner.phpp.components import (add_new_baseline_window_glazing, 
            add_new_baseline_window_frame, create_new_baseline_window_glazing)


def get_baseline_u_value(_baseline_code: BaselineCode, _climate_zone: ClimateZones) -> float:
    """Get the baseline U-Value for the specified climate zone."""
    u_values = _baseline_code.tables.fenestration_u_factors.operable_windows.u_factors
    u_values = getattr(u_values, _climate_zone.name)
    return u_values.group_r

def get_baseline_SHGC(_baseline_code: BaselineCode, _climate_zone: ClimateZones) -> float:
    """Get the baseline SHGC for the specified climate zone."""
    shgcs = _baseline_code.tables.fenestration_u_factors.operable_windows.shgc
    shgcs = getattr(shgcs, _climate_zone.name)
    return shgcs.group_r

def set_baseline_window_construction(
        _phpp_conn: phpp_app.PHPPConnection, 
        _baseline_code: BaselineCode,
        _climate_zone: ClimateZones,
) -> None:
    """Set the baseline window construction in the PHPP Windows Worksheet.
    
    Arguments:
    -----
        phpp_conn: phpp_app.PHPPConnection
            The PHPPConnection object
        baseline_code: BaselineCode
            The BaselineCode object
    """
    
    # -- Get the baseline values for U-Value and SHGC
    u_value = get_baseline_u_value(_baseline_code, _climate_zone)
    shgc =  get_baseline_SHGC(_baseline_code, _climate_zone)

    # -- Create the new baseline window construction
    baseline_phx_window = create_new_baseline_window_glazing(u_value, shgc)
    phpp_glazing_id = add_new_baseline_window_glazing(_phpp_conn, baseline_phx_window)
    phpp_frame_id = add_new_baseline_window_frame(_phpp_conn, baseline_phx_window)

    # -- Set the window constructions in the Windows Worksheet
    for row_num in _phpp_conn.windows.used_window_row_numbers:
        _phpp_conn.windows.set_single_window_construction_ids(row_num, phpp_glazing_id, phpp_frame_id)

def get_baseline_max_wwr(_baseline_code: BaselineCode) -> float:
    """Get the maximum window-to-wall ratio from the baseline code."""
    return _baseline_code.sections.maximum_fenestration_area.maximum_window_percent_of_wall

def get_baseline_max_srr(_baseline_code: BaselineCode) -> float:
    """Get the maximum skylight-to-roof ratio from the baseline code."""
    return _baseline_code.sections.maximum_fenestration_area.maximum_skylight_percent_of_roof

def set_baseline_window_area(_phpp_conn: phpp_app.PHPPConnection, _baseline_code: BaselineCode) -> None:
    """Set the maximum window-to-wall ratio in the PHPP Windows Worksheet.
    
    Arguments:
    ----------
        phpp_conn: phpp_app.PHPPConnection
            The PHPPConnection object
        baseline_code: BaselineCode
            The BaselineCode object
    """
    maximum_wwr = get_baseline_max_wwr(_baseline_code)
    current_wall_area = _phpp_conn.areas.get_total_wall_area()
    current_window_area = _phpp_conn.windows.get_total_window_area()
    current_wwr = current_window_area / (current_wall_area + current_window_area)

    if current_wwr < maximum_wwr:
        print(f"Current WWR {current_wwr:.2%} is less than maximum {maximum_wwr:.0%}.")
        return None

    print(f"Current WWR {current_wwr:.2%} is greater than maximum {maximum_wwr:.0%}. Scaling windows.")
    
    # -- Figure out the right scale factor
    scale_factor = maximum_wwr / current_wwr

    # -- Scale the window areas
    for row_num in _phpp_conn.windows.used_window_row_numbers:
        if not _phpp_conn.windows.row_is_window(row_num):
            continue

        _phpp_conn.windows.scale_window_size(row_num, scale_factor)


def set_baseline_skylight_area(_phpp_conn: phpp_app.PHPPConnection, _baseline_code: BaselineCode) -> None:
    """Set the maximum skylight-to-roof ratio in the PHPP Windows Worksheet.
    
    Arguments:
    ----------
        phpp_conn: phpp_app.PHPPConnection
            The PHPPConnection object
        baseline_code: BaselineCode
            The BaselineCode object
    """
    maximum_srr = get_baseline_max_srr(_baseline_code)
    current_roof_area = _phpp_conn.areas.get_total_roof_area()
    current_skylight_area = _phpp_conn.windows.get_total_skylight_area()
    current_srr = current_skylight_area / (current_roof_area + current_skylight_area)

    if current_srr < maximum_srr:
        print(f"Current SRR {current_srr:.2%} is less than maximum {maximum_srr:.0%}.")
        return None

    print(f"Current SRR {current_srr:.2%} is greater than maximum {maximum_srr:.0%}. Scaling Skylights.")
    
    # -- Figure out the right scale factor
    scale_factor = maximum_srr / current_srr

    # -- Scale the window areas
    for row_num in _phpp_conn.windows.used_window_row_numbers:
        if not _phpp_conn.windows.row_is_skylight(row_num):
            continue

        _phpp_conn.windows.scale_window_size(row_num, scale_factor) 



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Data model of the Baseline Building Code."""

from pydantic import BaseModel, validator
from typing import Dict

from ph_units.converter import convert


# -----------------------------------------------------------------------------

class TableMaximumUFactor_Values(BaseModel):
    group_r: float
    all_other: float

    @classmethod
    def convert_units(cls, v: "TableMaximumUFactor_Values", **kwargs):
        for field in v.__fields__:
            val = getattr(v, field)
            val = convert(val, kwargs["units"], "W/m2K")
            setattr(v, field, val)
        return v


class TableMaximumCFactor_Values(BaseModel):
    group_r: float
    all_other: float
    
    @classmethod
    def convert_units(cls, v: "TableMaximumCFactor_Values", **kwargs):
        for field in v.__fields__:
            val = getattr(v, field)
            val = convert(val, kwargs["units"], "W/m2K")
            setattr(v, field, val)
        return v
    

class TableMaximumFFactor_Values(BaseModel):
    group_r: float
    all_other: float
    
    @classmethod
    def convert_units(cls, v: "TableMaximumFFactor_Values", **kwargs):
        for field in v.__fields__:
            val = getattr(v, field)
            val = convert(val, kwargs["units"], "W/m2K")
            setattr(v, field, val)
        return v


class TableMaximumSHGC_Values(BaseModel):
    group_r: float
    all_other: float

    @classmethod
    def convert_units(cls, v: "TableMaximumSHGC_Values", **kwargs):
        return v
    

# -----------------------------------------------------------------------------


class TableMaximumUFactor_ClimateZones(BaseModel):
    CZ4: TableMaximumUFactor_Values 
    CZ5: TableMaximumUFactor_Values
    CZ6: TableMaximumUFactor_Values

    @classmethod
    def convert_units(cls, v: "TableMaximumUFactor_ClimateZones", **kwargs):
        for field_name in v.__fields__:
            field_type = getattr(v, field_name)
            field_type.convert_units(field_type, **kwargs)
        return v


class TableMaximumCFactor_ClimateZones(BaseModel):
    CZ4: TableMaximumCFactor_Values 
    CZ5: TableMaximumCFactor_Values
    CZ6: TableMaximumCFactor_Values

    @classmethod
    def convert_units(cls, v: "TableMaximumCFactor_ClimateZones", **kwargs):
        for field_name in v.__fields__:
            field_type = getattr(v, field_name)
            field_type.convert_units(field_type, **kwargs)
        return v


class TableMaximumFFactor_ClimateZones(BaseModel):
    CZ4: TableMaximumFFactor_Values
    CZ5: TableMaximumFFactor_Values
    CZ6: TableMaximumFFactor_Values
    
    @classmethod
    def convert_units(cls, v: "TableMaximumFFactor_ClimateZones", **kwargs):
        for field_name in v.__fields__:
            field_type = getattr(v, field_name)
            field_type.convert_units(field_type, **kwargs)
        return v


class TableMaximumSHGC_ClimateZones(BaseModel):
    CZ4: TableMaximumSHGC_Values
    CZ5: TableMaximumSHGC_Values
    CZ6: TableMaximumSHGC_Values
    
    @classmethod
    def convert_units(cls, v: "TableMaximumSHGC_ClimateZones", **kwargs):
        return v


# -----------------------------------------------------------------------------


class TableMaximumUFactorRoofs(BaseModel):
    insulation_above_deck: TableMaximumUFactor_ClimateZones
    metal_building: TableMaximumUFactor_ClimateZones
    attic_and_other: TableMaximumUFactor_ClimateZones

    @classmethod
    def convert_units(cls, v: "TableMaximumUFactorRoofs", **kwargs):
        for field_name in v.__fields__:
            field_type = getattr(v, field_name)
            field_type.convert_units(field_type, **kwargs)
        return v


class TableMaximumUFactorWalls(BaseModel):
    mass: TableMaximumUFactor_ClimateZones
    metal_building: TableMaximumUFactor_ClimateZones
    metal_framed: TableMaximumUFactor_ClimateZones
    wood_framed_and_other: TableMaximumUFactor_ClimateZones
    below_grade: TableMaximumCFactor_ClimateZones

    @classmethod
    def convert_units(cls, v: "TableMaximumUFactorWalls", **kwargs):
        for field_name in v.__fields__:
            field_type = getattr(v, field_name)
            field_type.convert_units(field_type, **kwargs)
        return v


class TableMaximumUFactorFloors(BaseModel):
    mass: TableMaximumUFactor_ClimateZones
    joist_framed: TableMaximumUFactor_ClimateZones
    unheated_slab: TableMaximumFFactor_ClimateZones
    heated_slab: TableMaximumUFactor_ClimateZones

    @classmethod
    def convert_units(cls, v: "TableMaximumUFactorFloors", **kwargs):
        for field_name in v.__fields__:
            field_type = getattr(v, field_name)
            field_type.convert_units(field_type, **kwargs)
        return v


class TableMaximumUFactorFenestration(BaseModel):
    u_factors: TableMaximumUFactor_ClimateZones
    shgc: TableMaximumSHGC_ClimateZones
    
    @classmethod
    def convert_units(cls, v: "TableMaximumUFactorFenestration", **kwargs):
        for field_name in v.__fields__:
            field_type = getattr(v, field_name)
            field_type.convert_units(field_type, **kwargs)
        return v


# -----------------------------------------------------------------------------


class TableEnvelopeMaximumUFactors(BaseModel):
    name: str
    section: str
    units: str
    roofs: TableMaximumUFactorRoofs
    walls: TableMaximumUFactorWalls
    floors: TableMaximumUFactorFloors

    @validator("roofs", "walls", "floors", pre=False)
    @classmethod
    def convert_units(cls, v: "TableEnvelopeMaximumUFactors", values: Dict):
        return type(v).convert_units(v, **values)


class TableFenestrationMaximumUFactors(BaseModel):  
    name: str
    section: str
    units: str
    fixed_windows: TableMaximumUFactorFenestration
    operable_windows: TableMaximumUFactorFenestration
    entrance_doors: TableMaximumUFactorFenestration
    skylights: TableMaximumUFactorFenestration
    
    @validator("fixed_windows", "operable_windows", "entrance_doors", "skylights", pre=False)
    @classmethod
    def convert_units(cls, v: "TableFenestrationMaximumUFactors", values: Dict):
        return type(v).convert_units(v, **values)


class TableLightingLPDAreaMethod(BaseModel):
    name: str
    section: str
    units: str
    LPD: dict
    
    @validator("LPD", pre=False)
    @classmethod
    def convert_units(cls, v: Dict, values: Dict):
        units = values["units"]
        for key, val in v.items():
            v[key] = convert(val, units, "W/m2")
        return v


# -----------------------------------------------------------------------------


class SectionMaximumFenestrationArea(BaseModel):
    source_url: str
    name: str
    section: str
    maximum_skylight_percent_of_roof: float
    maximum_window_percent_of_wall: float
    

# -----------------------------------------------------------------------------


class BaselineCodeTables(BaseModel):
    maximum_u_factors: TableEnvelopeMaximumUFactors
    fenestration_u_factors: TableFenestrationMaximumUFactors
    lighting_area_method: TableLightingLPDAreaMethod


class BaselineCodeSections(BaseModel):
    maximum_fenestration_area :SectionMaximumFenestrationArea


class BaselineCode(BaseModel):
    """Pydantic BaselineCode model."""
    name: str
    state: str
    year: str
    source_url: str
    tables: BaselineCodeTables
    sections: BaselineCodeSections



# -*- coding: utf-8 -*-
# -*- Python Version: 3.7 -*-

"""Functions for creating and setting Baseline Constructions on the PHPP U-Values worksheet."""

from copy import copy
from typing import NamedTuple

from PHX.PHPP import phpp_app
from PHX.model.constructions import PhxConstructionOpaque

from ph_baseliner.codes.model import BaselineCode
from ph_baseliner.codes.options import ClimateZones


class BaselinePHXConstructions(NamedTuple):
    """NamedTuple: The baseline PHX constructions for the PHPP U-Values Worksheet."""
    roof: PhxConstructionOpaque
    exposed_wall: PhxConstructionOpaque
    ground_wall: PhxConstructionOpaque
    exposed_floor: PhxConstructionOpaque
    ground_floor: PhxConstructionOpaque


class BaselineConstructionPHPPids(NamedTuple):
    """NamedTuple: Output of the add_baseline_constructions_to_phpp function."""
    roof: str
    exposed_wall: str
    ground_wall: str
    exposed_floor: str
    ground_floor: str


def create_baseline_constructions(
        baseline_code: BaselineCode,
        climate_zone: ClimateZones,
) -> BaselinePHXConstructions:
    """Create the baseline constructions from the code baseline model."""
    table_max_u_values = baseline_code.tables.maximum_u_factors
    
    roof_u_values = getattr(table_max_u_values.roofs.insulation_above_deck, climate_zone.name)
    new_roof_phx_construction = PhxConstructionOpaque.from_total_u_value(
        roof_u_values.group_r, "BASELINE: ROOF"
    )

    exposed_wall_u_values = getattr(table_max_u_values.walls.mass, climate_zone.name)
    new_exposed_wall_phx_construction = PhxConstructionOpaque.from_total_u_value(
        exposed_wall_u_values.group_r, "BASELINE: EXPOSED WALL"
    )

    ground_wall_u_values = getattr(table_max_u_values.walls.below_grade, climate_zone.name)
    new_ground_wall_phx_construction = PhxConstructionOpaque.from_total_u_value(
        ground_wall_u_values.group_r, "BASELINE: GROUND WALL"
    )

    exposed_floor_u_values = getattr(table_max_u_values.floors.mass, climate_zone.name)
    new_exposed_floor_phx_construction = PhxConstructionOpaque.from_total_u_value(
        exposed_floor_u_values.group_r, "BASELINE: EXPOSED FLOOR"
    )
    
    ground_floor_u_values = getattr(table_max_u_values.floors.unheated_slab, climate_zone.name)
    new_ground_floor_phx_construction = PhxConstructionOpaque.from_total_u_value(
        ground_floor_u_values.group_r,"BASELINE: GROUND FLOOR"
    )

    return BaselinePHXConstructions(
        new_roof_phx_construction,
        new_exposed_wall_phx_construction,
        new_ground_wall_phx_construction,
        new_exposed_floor_phx_construction,
        new_ground_floor_phx_construction,
    )

def add_baseline_constructions_to_phpp(phpp_conn: phpp_app.PHPPConnection, baseline_constructions: BaselinePHXConstructions) -> BaselineConstructionPHPPids:
    """Add the baseline constructions to the PHPP U-Values Worksheet."""
    roof_phpp_id = phpp_conn.u_values.add_new_phx_construction(baseline_constructions.roof)
    exposed_wall_phpp_id = phpp_conn.u_values.add_new_phx_construction(baseline_constructions.exposed_wall)
    ground_wall_phpp_id = phpp_conn.u_values.add_new_phx_construction(baseline_constructions.ground_wall)
    exposed_floor_phpp_id = phpp_conn.u_values.add_new_phx_construction(baseline_constructions.exposed_floor)
    ground_floor_phpp_id = phpp_conn.u_values.add_new_phx_construction(baseline_constructions.ground_floor)

    return BaselineConstructionPHPPids(roof_phpp_id, exposed_wall_phpp_id, ground_wall_phpp_id, exposed_floor_phpp_id, ground_floor_phpp_id)

def replace_u_values_with_baseline_constructions(phpp_conn: phpp_app.PHPPConnection, baseline_constructions: BaselinePHXConstructions) -> None:
    """Replace the existing U-Values with the baseline constructions."""

    def is_roof(_input: str):
        """Return True if the R_si type is a 'Roof'."""
        return "ROOF" in str(_input).upper()

    def is_wall(_input: str):
        """Return True if the R_si type is a 'Wall'."""
        return "WALL" in str(_input).upper()

    def is_floor(_input: str):
        """Return True if the R_si type is a 'Floor'."""
        return "FLOOR" in str(_input).upper()

    def is_below_grade(_input: str):
        """Return True if the R_se type is a 'Ground'."""
        return "GROUND" in str(_input).upper()

    for row_num in phpp_conn.u_values.used_constructor_start_rows:
        constructor_name = phpp_conn.u_values.get_constructor_name(row_num)
        r_si_type = phpp_conn.u_values.get_constructor_r_si_type(row_num)
        r_se_type = phpp_conn.u_values.get_constructor_r_se_type(row_num)
        
        if is_roof(r_si_type):
            _phx_const = copy(baseline_constructions.roof)
        elif is_wall(r_si_type) and is_below_grade(r_se_type):
            _phx_const = copy(baseline_constructions.ground_wall)
        elif is_wall(r_si_type) and not is_below_grade(r_se_type):
            _phx_const = copy(baseline_constructions.exposed_wall)
        elif is_floor(r_si_type) and is_below_grade(r_se_type):
            _phx_const = copy(baseline_constructions.ground_floor)
        elif is_floor(r_si_type) and not is_below_grade(r_se_type):
            _phx_const = copy(baseline_constructions.exposed_floor)
        else:
            continue

        _phx_const.display_name = constructor_name
        
        # -- Update the PHPP
        phpp_conn.u_values.clear_single_constructor_data(row_num, False)
        phpp_conn.u_values.write_single_PHX_construction(_phx_const, row_num)